#!/usr/bin/env python3
# Copyright 2023 Jason C. Nucciarone
# See LICENSE file for licensing details.

"""Public utilities used inside cleantest.

You can use these utilities for developing third-party plugins.
"""

from .detect_os import detect_os_variant
from .thread_count import thread_count
